import logging
from enum import Enum

from kombu import Connection, Consumer, Exchange, Queue
from kombu.pools import producers
from kombu.utils import nested

import ebosia
import ebosia.connection
import ebosia.events

LOGGER = logging.getLogger(__name__)

def errback(exc, interval):
    LOGGER.error("Failed to publish message: %r", exc, exc_info=1)
    LOGGER.info("Retrying publish in %s seconds", interval)

class ConnectionSync(ebosia.connection.ConnectionBase):
    def __init__(self, uri, timeout=None):
        self.consumers = []
        self.timeout = timeout
        self._connection = Connection(uri)
        self._connection.connect()

    def publish(self, topic, payload):
        LOGGER.debug("Publishing %s %s", topic, payload)
        exchange = Exchange('eventbus', type='topic')
        with producers[self._connection].acquire(block=True) as producer:
            _publish = self._connection.ensure(producer, producer.publish, errback=errback, max_retries=3)
            return _publish(
                payload,
                serializer  = 'json',
                compression = 'bzip2',
                exchange    = exchange,
                routing_key = topic if not isinstance(topic, Enum) else str(topic),
            )

    def subscribe(self, routing_key, callback):
        def _on_message(body, message):
            event = ebosia.events.from_kombu(body, message)
            callback(event)
        exchange = Exchange('eventbus', type='topic', durable=False)
        queue = Queue('', exchange, routing_key=routing_key)
        consumer = Consumer(self._connection, [queue], callbacks=[_on_message])
        self.consumers.append(consumer)

    def drain(self):
        with nested(*self.consumers):
            self._connection.drain_events(timeout=self.timeout)

def connect(uri):
    connection = ConnectionSync(uri)
    ebosia.store(connection)
    return connection

def publish(topic, payload, bus=None):
    bus = bus or ebosia.get()
    return bus.publish(topic, payload)

def subscribe(routing_key, callback, bus=None):
    bus = bus or ebosia.get()
    return bus.subscribe(routing_key, callback)

def drain(bus=None):
    bus = bus or ebosia.get()
    return bus.drain()



import bz2
import collections
import json

Event = collections.namedtuple('Event', ['topic', 'payload'])

def from_kombu(body, message):
    #topic = message.properties['delivery_info'].get('routing_key')
    topic = message.delivery_info['routing_key']
    return Event(topic, body)

def from_aioamqp(body, envelope, properties): # pylint: disable=unused-argument
    decompressed = bz2.decompress(body)
    decoded = decompressed.decode('utf-8')
    parsed = json.loads(decoded)
    topic = envelope.routing_key
    return Event(topic, parsed)



import collections

import pytest

import ebosia as _ebosia
from ebosia import sync


class TestConnection(sync.ConnectionSync):
    PRETENDING_TO_BE_ASYNC = True
    def __init__(self):
        super().__init__('memory:///', 0.1)
        self.events         = collections.deque()
        self.consumer       = super().subscribe('#', self.process)
        self.subscriptions  = []

    def process(self, event):
        self.events.append(event)
        for topic, callback in self.subscriptions:
            if topic == '#' or event.topic == topic:
                callback(event)

    def publish(self, topic, payload):
        super().publish(topic, payload)
        self.drain()

    def published(self, topic, payload=None):
        if len(self.events) < 1:
            return False
        event = self.events.popleft()
        if payload is None:
            return event.topic == topic
        else:
            return event.topic == topic and event.payload == payload

    def subscribe(self, topic, callback):
        self.subscriptions.append((topic, callback))

@pytest.yield_fixture
def eventbus_connection():
    previous = None
    try:
        previous = _ebosia.get()
    except Exception: # pylint: disable=broad-except
        pass
    connection = TestConnection()
    _ebosia.store(connection)
    yield connection
    _ebosia.store(previous)

@pytest.yield_fixture
def eventbus(eventbus_connection): # pylint: disable=redefined-outer-name
    yield eventbus_connection



#!/usr/bin/env python
import contextlib
import imp
import os
import re

from setuptools import setup


def _git_to_version(git):
    match = re.match(r'(?P<tag>[\d\.]+)-(?P<offset>[\d]+)-(?P<sha>\w{8})', git)
    if not match:
        version = git
        branch = None
    else:
        branch = os.environ.get('GIT_BRANCH', None)
        version = "{tag}.post{offset}".format(**match.groupdict())
    print("Calculated ebosia version '{}' from git description '{}' and branch '{}'".format(version, git, branch))
    return version

@contextlib.contextmanager
def write_version():
    VERSION_FILE = 'ebosia/version.py'
    git_description = os.environ.get('GITDESCRIPTION', None)
    version = _git_to_version(git_description) if git_description else None
    if version:
        with open(VERSION_FILE, 'r') as f:
            old_contents = f.read()
        with open(VERSION_FILE, 'w') as f:
            f.write('VERSION = "{}"\n'.format(version))
    yield
    if version:
        with open(VERSION_FILE, 'w') as f:
            f.write(old_contents)

def get_version():
    basedir = os.path.abspath(os.path.dirname(__file__))
    version = imp.load_source('version', os.path.join(basedir, 'ebosia', 'version.py'))
    return version.VERSION

def main():
    with write_version():
        setup(
            name                = 'ebosia',
            version             = get_version(),
            description         = "A library for getting events throughout the system",
            long_description    = open('README.md').read(),
            author              = 'Authentise, Inc.',
            author_email        = 'engineering@authentise.com',
            entry_points        = {
                'pytest11'      : [
                    'ebosia = ebosia.fixtures'
                ]
            },
            install_requires    = [
                'aioamqp==0.6.0',
                'kombu==3.0.34', # sepiida.celery uses kombu 3.0.34
            ],
            extras_require      = {
                'develop'       : [
                    'coverage==3.7.1',
                    'mothermayi==0.4',
                    'mothermayi-pylint==0.5',
                    'mothermayi-isort==0.5',
                    'pylint==1.5.2',
                    'pytest==2.8.5',
                    'pytest-asyncio==0.1.3',
                    'pytest-cov==1.8.1',
                ]
            },
            packages            = ['ebosia'],
            package_data        = {
                'ebosia'      : ['ebosia/*'],
            },
            scripts             = ['bin/sync-emitter', 'bin/sync-subscribe', 'bin/async-subscribe'],
            include_package_data= True,
        )

if __name__ == '__main__':
    main()



import asyncio
import logging
import urllib.parse

import aioamqp

import ebosia
import ebosia.events

LOGGER = logging.getLogger(__name__)

class ClosedConnection(Exception):
    pass

class Connection(): # pylint: disable=too-many-instance-attributes
    def __init__(self, url, on_error=None, on_reconnect=None, heartbeat=60):
        parts = urllib.parse.urlparse(url)
        if parts.scheme != 'amqp':
            raise Exception("Only 'amqp' is supported for the connection scheme at this time")

        self.channel        = None
        self.heartbeat      = heartbeat
        self.on_error       = on_error
        self.on_reconnect   = on_reconnect
        self.parts          = parts
        self.protocol       = None
        self.reconnect_lock = asyncio.Lock()
        self.transport      = None


    @asyncio.coroutine
    def error_callback(self, error):
        LOGGER.warning("Error in connection to eventbus: %s", error)
        if self.on_error:
            yield from self.on_error(error)
        if self.reconnect_lock.locked():
            LOGGER.debug("Not reconnecting, the reconnect lock is engaged")
        else:
            asyncio.async(self.reconnect()) # pylint: disable=deprecated-method


    @asyncio.coroutine
    def reconnect(self):
        with (yield from self.reconnect_lock):
            wait_time = 0.1
            while True:
                try:
                    yield from self.connect()
                    if self.on_reconnect:
                        yield from self.on_reconnect()
                    return
                except OSError as e:
                    LOGGER.warning("Unable to reconnect: %s", e)
                    yield from asyncio.sleep(wait_time)
                    wait_time = min(3.0, wait_time * 1.5)

    @asyncio.coroutine
    def connect(self):
        try:
            transport, protocol = yield from aioamqp.connect(
                host        = self.parts.hostname,
                port        = self.parts.port,
                login       = self.parts.username,
                password    = self.parts.password,
                virtualhost = self.parts.path or '/',
                insist      = True,
                on_error    = self.error_callback,
                heartbeat   = self.heartbeat,
            )
        except aioamqp.AmqpClosedConnection:
            raise ClosedConnection

        self.channel   = yield from protocol.channel()
        self.protocol  = protocol
        self.transport = transport

@asyncio.coroutine
def declare_exchange(connection=None):
    connection = connection or ebosia.get()
    exchange = yield from connection.channel.exchange(
        'eventbus',
        'topic',
        passive         = False,
        durable         = False,
        auto_delete     = False,
        no_wait         = False,
    )
    return exchange

def _sync_to_async(callback):
    def _on_done(result):
        if result.exception():
            raise result.exception()

    def _sync_callback(event):
        loop = asyncio.get_event_loop()
        coro = loop.create_task(callback(event))
        coro.add_done_callback(_on_done)
        loop.run_until_complete(coro)
    return _sync_callback

@asyncio.coroutine
def subscribe(topic, callback, connection=None):
    connection = connection or ebosia.get()
    # If we are testing the connection we receive will actually be an in-memory
    # kombu connection. That's fine, we just need to do some translation
    if getattr(connection, 'PRETENDING_TO_BE_ASYNC', None):
        return connection.subscribe(topic, _sync_to_async(callback))

    @asyncio.coroutine
    def on_message(channel, body, envelope, properties): # pylint: disable=unused-argument
        event = ebosia.events.from_aioamqp(body, envelope, properties)
        LOGGER.debug("Received %s", event)
        yield from callback(event)

    yield from declare_exchange(connection)
    result = yield from connection.channel.queue_declare(
        queue_name  = '',
        passive     = False,
        durable     = False,
        exclusive   = True,
        no_wait     = False,
    )
    yield from connection.channel.queue_bind(
        queue_name      = result['queue'],
        exchange_name   = 'eventbus',
        routing_key     = topic,
        no_wait         = False,
    )
    yield from connection.channel.basic_consume(on_message, queue_name=result['queue'], no_ack=True)

@asyncio.coroutine
def connect(url, on_error=None, on_reconnect=None, heartbeat=60):
    connection = Connection(url, on_error, on_reconnect=on_reconnect, heartbeat=heartbeat)
    yield from connection.connect()
    ebosia.store(connection)
    return connection

@asyncio.coroutine
def disconnect(connection=None):
    connection = connection or ebosia.get()
    yield from connection.protocol.close()
    connection.transport.close()
